/*
 *     DroidScrape
 *     Copyright (C) 2017  Mario de Benito
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Also add information on how to contact you by electronic and paper mail.
 *
 *   If the program does terminal interaction, make it output a short
 * notice like this when it starts in an interactive mode:
 *
 *     DroidScrape  Copyright (C) 2017  Mario de Benito
 *     This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
 *     This is free software, and you are welcome to redistribute it
 *     under certain conditions; type `show c' for details.
 *
 * The hypothetical commands `show w' and `show c' should show the appropriate
 * parts of the General Public License.  Of course, your program's commands
 * might be different; for a GUI interface, you would use an "about box".
 *
 *   You should also get your employer (if you work as a programmer) or school,
 * if any, to sign a "copyright disclaimer" for the program, if necessary.
 * For more information on this, and how to apply and follow the GNU GPL, see
 * <http://www.gnu.org/licenses/>.
 *
 *   The GNU General Public License does not permit incorporating your program
 * into proprietary programs.  If your program is a subroutine library, you
 * may consider it more useful to permit linking proprietary applications with
 * the library.  If this is what you want to do, use the GNU Lesser General
 * Public License instead of this License.  But first, please read
 * <http://www.gnu.org/philosophy/why-not-lgpl.html>.
 */

package mdbdev.es.droidscrape.plugins.emailheaderanalyzer.analyzer;

import android.support.annotation.NonNull;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.mail.Header;
import javax.mail.MessagingException;
import javax.mail.internet.InternetHeaders;

import mdbdev.es.droidscrape.plugins.emailheaderanalyzer.analyzer.ipapi.IpApi;
import mdbdev.es.droidscrape.plugins.emailheaderanalyzer.analyzer.model.AnalyzedHeader;
import mdbdev.es.droidscrape.plugins.emailheaderanalyzer.analyzer.model.EmailHeadersReport;


/**
 * Created by Mario on 22/04/2017.
 * Using: http://ip-api.com/docs/api:json
 * Class that performs the header analysis
 */
public class EmailHeaderAnalyzer {
    private List<EmailHeaderAnalysisListener> listeners;

    public EmailHeaderAnalyzer() {
        this.listeners = new ArrayList<>();
    }

    /**
     * Analyses the email headers contained in the string received
     * @param string Email headers string
     * @return Result of the analysis
     * @throws UnsupportedEncodingException
     * @throws MessagingException
     */
    public EmailHeadersReport analyze(String string) throws UnsupportedEncodingException, MessagingException {

        IpApi.START_TIMESTAMP = new Date().getTime();

        EmailHeadersReport result = new EmailHeadersReport();
        //Make sure we have the correct "end of header" signal, a line break
        string +=System.getProperty("line.separator");
        //Load the header text into a InternetHeaders object for easier processing
        InputStream stream = new ByteArrayInputStream(string.getBytes("UTF-8"));
        InternetHeaders ih = new InternetHeaders();
        ih.load(stream);
        //Iterate the headers and process them one by one
        Enumeration<Header> headers = ih.getAllHeaders();
        int count = 0;
        while (headers.hasMoreElements()){
            Header h = headers.nextElement();
            count++;
            publishProgressUpdate(h,count);

            String analysis = "";
            if(validateHeader(h))
                analysis = analyzeHeaderValue(h.getValue().replace("\n", " ").replace("\r", " "));
            else
                analysis = "Invalid header.";
            AnalyzedHeader ah = new AnalyzedHeader(h.getName(),h.getValue(),analysis);
            result.addHeader(ah);
        }
        IpApi.START_TIMESTAMP = 0;
        return result;
    }

    /**
     * Notifies the listeners about progress in the analysis
     * @param h Header being analysed
     * @param count
     */
    private void publishProgressUpdate(Header h, int count) {
        for(EmailHeaderAnalysisListener l : listeners)
            l.updateProgress("Analyzing header: ["+count+"] "+h.getName());
    }

    /**
     * Analyses a single header
     * @param h Header value string
     * @return
     */
    private static String analyzeHeaderValue(String h)  {
        String analysis = "<dl>";

        analysis += checkIPAddresses(h);
        analysis += checkDomains(h);

        analysis += "</dl>";
        return analysis;
    }

    /**
     * Checks the string for existing domains and analyses them
     * @param h Header value string
     * @return Domain analysis string
     */
    private static String checkDomains(String h)  {
        //Pattern pattern = Patterns.DOMAIN_NAME;
        // Standard pattern hangs the device due to weird domain-like strings present in many headers.
        // Using a custom one
        Pattern pattern = Pattern.compile("(([a-z]|[A-Z]|[0-9]|-){1,63}\\.)+(([a-z]|[A-Z]){2,8})");
        Matcher matcher = pattern.matcher(h);
        List<String> domains = new ArrayList<>();
        while (matcher.find()) {
            String domain = matcher.group();
            if(!domains.contains(domain))
                domains.add(domain);
        }
        String analysis = "";
        if(domains.size()>0)
            analysis += "<b>Domains:</b> "+System.getProperty("line.separator");
        for(String domain : domains){
            analysis += analyzeDomain(domain);
        }
        return analysis;
    }

    /**
     * Analyses a domain
     * @param domain Domain to analyse
     * @return Domain analysis string
     */
    private static String analyzeDomain(String domain) {
        String analysis = "<dt>"+domain+"</dt>";
        InetAddress addr = null;
        try {
            addr = InetAddress.getByName(domain);
            analysis+="<dd>IP: "+addr.getHostAddress()+"; ";
            try {
                analysis = queryForIP(addr.getHostAddress(), analysis);
            } catch (IOException e) {
                //e.printStackTrace();
                analysis+="IP details could not be retrieved.";
            } catch (JSONException e) {
                //e.printStackTrace();
                analysis+="IP details could not be retrieved.";
            }
            analysis+="</dd>";
        } catch (UnknownHostException e) {
            //e.printStackTrace();
            analysis+="<dd>Could not retrieve information about the host.</dd>";
        }
        return analysis+System.getProperty("line.separator");
    }

    /**
     * Checks the string for existing IP addresses and analyses them
     * @param h Header value string
     * @return Header analysis string
     */
    private static String checkIPAddresses(String h)  {
        //Pattern pattern = Patterns.IP_ADDRESS;
        //Using custom pattern to ensure we get the same results in any Android version
        Pattern pattern = Pattern.compile("((25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[1-9][0-9]|[1-9])\\.(25[0-5]|2[0-4]"
                + "[0-9]|[0-1][0-9]{2}|[1-9][0-9]|[1-9]|0)\\.(25[0-5]|2[0-4][0-9]|[0-1]"
                + "[0-9]{2}|[1-9][0-9]|[1-9]|0)\\.(25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}"
                + "|[1-9][0-9]|[0-9]))");
        Matcher matcher = pattern.matcher(h);
        List<String> IPAddresses = new ArrayList<>();
        while (matcher.find()) {
            String ipAddress = matcher.group();
            if(!IPAddresses.contains(ipAddress))
                IPAddresses.add(ipAddress);
        }
        String analysis = "";
        if(IPAddresses.size()>0)
            analysis += "<b>IP Addresses:</b> "+System.getProperty("line.separator");
        for(String ip : IPAddresses){
            analysis += analyzeIPAddress(ip);
        }
        return analysis;
    }

    /**
     * Analyses an IP address
     * @param ipAddress IP address to analyse
     * @return IP Address analysis
     */
    private static String analyzeIPAddress(String ipAddress)  {
        String analysis = "<dt>"+ipAddress+"</dt>";
        InetAddress addr = null;
        try {
            addr = InetAddress.getByName(ipAddress);
            boolean privateAddr = addr.isSiteLocalAddress();
            if(privateAddr) {
                analysis += "<dd>Private</dd>";
            }else{
                analysis +="<dd>";

                try {
                    analysis = queryForIP(ipAddress, analysis);
                } catch (IOException e) {
                    //e.printStackTrace();
                    analysis+="IP details could not be retrieved.";
                } catch (JSONException e) {
                    //e.printStackTrace();
                    analysis+="IP details could not be retrieved.";
                }
                analysis+="</dd>";
            }
        } catch (UnknownHostException e) {
            e.printStackTrace();
            analysis+="<dd>Could not retrieve information about the host.</dd>";
        }
        return analysis+System.getProperty("line.separator");
    }

    /**
     * Queries IP Api for the received IP address and adds the information gathered to the analysis string received
     * @param ipAddress IP address to query
     * @param analysis Analysis string to complete
     * @return Completed analysis string
     * @throws IOException
     * @throws JSONException
     */
    @NonNull
    private static String queryForIP(String ipAddress, String analysis) throws IOException, JSONException {
        JSONObject ipApi = IpApi.queryForIp(ipAddress);
        String status = ipApi.getString("status");
        if(status.equalsIgnoreCase("success")) {
            String details = ipApi.getString("city") + "(" + ipApi.getString("countryCode") + "); Org: " + ipApi.getString("org") + "; ISP: " + ipApi.getString("isp");
            analysis += details;
        }else{
            throw new IOException("Unsuccessful API call.");
        }
        return analysis;
    }

    /**
     * Checks if the header provided is valid (syntactically)
     * @return True if valid
     * @param h Header to be validated
     */
    private boolean validateHeader(Header h) {
        Pattern pattern = Pattern.compile("^[a-zA-Z0-9-]+$");
        Matcher nameMatcher = pattern.matcher(h.getName());
        if(nameMatcher.find())
            return true;
        return false;
    }


    /**
     * Adds an event listener for the analysis project
     * @param l
     */
    public void addListener(EmailHeaderAnalysisListener l){
        listeners.add(l);
    }

}