/*
 *     DroidScrape
 *     Copyright (C) 2017  Mario de Benito
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Also add information on how to contact you by electronic and paper mail.
 *
 *   If the program does terminal interaction, make it output a short
 * notice like this when it starts in an interactive mode:
 *
 *     DroidScrape  Copyright (C) 2017  Mario de Benito
 *     This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
 *     This is free software, and you are welcome to redistribute it
 *     under certain conditions; type `show c' for details.
 *
 * The hypothetical commands `show w' and `show c' should show the appropriate
 * parts of the General Public License.  Of course, your program's commands
 * might be different; for a GUI interface, you would use an "about box".
 *
 *   You should also get your employer (if you work as a programmer) or school,
 * if any, to sign a "copyright disclaimer" for the program, if necessary.
 * For more information on this, and how to apply and follow the GNU GPL, see
 * <http://www.gnu.org/licenses/>.
 *
 *   The GNU General Public License does not permit incorporating your program
 * into proprietary programs.  If your program is a subroutine library, you
 * may consider it more useful to permit linking proprietary applications with
 * the library.  If this is what you want to do, use the GNU Lesser General
 * Public License instead of this License.  But first, please read
 * <http://www.gnu.org/philosophy/why-not-lgpl.html>.
 */

package mdbdev.es.droidscrape.plugins.webscraper.scraper;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import mdbdev.es.droidscrape.plugins.webscraper.scraper.model.Occurrence;
import mdbdev.es.droidscrape.plugins.webscraper.scraper.model.ScraperLink;
import mdbdev.es.droidscrape.plugins.webscraper.scraper.model.ScraperMode;
import mdbdev.es.droidscrape.plugins.webscraper.scraper.model.ScraperReport;

/**
 * Created by Mario on 29/04/2017.
 * Class that performs the scraping process for the Web Scraper module
 */
public class WebScraper {
    private String url;
    private String baseUrl;
    private String terms;
    private long depth;
    private ScraperMode mode;
    private List<ScraperStatusListener> listeners;
    private String HTMLReport;

    private LinkedList<ScraperLink> pendingLinks;
    private LinkedList<String> scannedLinks;
    private List<Occurrence> occurrences;
    private List<String> searchTerms;


    /**
     * Constructor that configures the scraper
     * @param url URL to startScraper
     * @param terms Terms to search in that site
     * @param depth Link depth to scan
     * @param mode Aggressiveness of the scan
     */
    public WebScraper(String url, String terms, long depth, ScraperMode mode) throws MalformedURLException {
        this.url = url;
        //Extract base URL
        URL tmp = new URL(url);
        this.baseUrl = tmp.getHost();
        this.terms = terms;
        this.depth = depth;
        this.mode = mode;
        this.HTMLReport = "<h1>Web Scraper report</h1>";
        this.listeners = new ArrayList<>();
        this.pendingLinks = new LinkedList<>();
        this.scannedLinks = new LinkedList<>();
        this.occurrences = new ArrayList<>();
        this.searchTerms = new ArrayList<>();

        String[] termsArr = terms.split(System.getProperty("line.separator"));
        searchTerms.addAll(Arrays.asList(termsArr));

        pendingLinks.push(new ScraperLink(url,0));
    }

    /**
     * Starts the scraper
     */
    public void startScraper(){
        scrape();
    }

    /**
     * Scrapes the next available URL
     */
    private void scrape() {
        //Get the first document of the list
        if (!pendingLinks.isEmpty()) {
            ScraperLink sl = pendingLinks.pop();
            String currentUrl = sl.getUrl();
            int currentDepth = sl.getDepth();
            publishCurrentUrlChanged(currentUrl,currentDepth);
            try {
                Document doc = Jsoup.connect(currentUrl).get();
                //Get the links if we still need them
                if (currentDepth < depth)
                    addAllLinks(doc,currentDepth+1);
                //Get occurrences in current document
                scanForOccurrences(doc,currentUrl);
                //Clean garbage
                doc=null;
            } catch (IOException e) {
                e.printStackTrace();
                publishUrlError(currentUrl,e.getMessage());
            }
            //Register the link as scanned
            scannedLinks.push(currentUrl);

            publishFinishedScanningUrl(currentUrl);
            //Wait for the configured time
            publishWaitStatus(true);

            try {
                Thread.sleep(this.mode.getWaitTime());
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            publishWaitStatus(false);
            //Proceed to next piled link
            scrape();
        }else{
            publishScrapeFinished();
        }

    }
    /**
     * Scans the document and adds all the occurrences of terms to the results list
     * @param doc Document object to be scanned
     * @param currentUrl
     */
    private void scanForOccurrences(Document doc, String currentUrl) {
        for(String term : searchTerms){
            Elements elms = doc.select("*:containsOwn("+term+")");
            if(elms.size()>0){
                for(Element e : elms) {
                    Occurrence o = new Occurrence(term, currentUrl, e.outerHtml());
                    occurrences.add(o);
                }
            }
        }

    }

    /**
     * Retrieves all found links in the site scanned
     * @param doc Document object of the site to scan
     * @param currentDepth
     */
    private void addAllLinks(Document doc, int currentDepth) throws IOException {
        Elements linkElms = doc.select("a[href]");
        int i;
        Element e;
        for(i=0;i<linkElms.size();i++){
            e = linkElms.get(i);
            String href= e.attr("abs:href");
            if(isValidLink(href)){
                addToLinkPool(href,currentDepth);
            }
        }
    }

    /**
     * Checks if a link is valid to be scanned:
     * -It is in the same domain we are scanning
     * -It is not a javascript:function like link
     * @param href URL to validate
     * @return
     */
    private boolean isValidLink(String href) throws MalformedURLException {
        if(href.startsWith("javascript:")){
            return false;
        }
        URL validatingUrl = new URL(href);
        String validatingHost = validatingUrl.getHost();
        if(!validatingHost.equals(baseUrl)){
            return false;
        }
        return true;
    }

    /**
     * Adds a link to the pending pool
     * @param href URL of the link
     * @param currentDepth
     */
    private void addToLinkPool(String href, int currentDepth) {
        if(!pendingLinks.contains(new ScraperLink(href,0)) &&
                !pendingLinks.contains(new ScraperLink(href+"/",0)) &&
                !scannedLinks.contains(href) &&
                !scannedLinks.contains(href+"/")) {
            pendingLinks.push(new ScraperLink(href,currentDepth));
            publishFoundNewLink(href,pendingLinks.size());
        }
    }

    /**
     * Adds a status listener that will receive the events generated by the scraper
     * @param listener Object that will receive the events
     */
    public void addStatusListener(ScraperStatusListener listener){
        this.listeners.add(listener);
    }

    public String getUrl() {
        return url;
    }

    public void setUrl(String url) {
        this.url = url;
    }

    public String getTerms() {
        return terms;
    }

    public void setTerms(String terms) {
        this.terms = terms;
    }

    public long getDepth() {
        return depth;
    }

    public void setDepth(long depth) {
        this.depth = depth;
    }

    public ScraperMode getMode() {
        return mode;
    }

    public void setMode(ScraperMode mode) {
        this.mode = mode;
    }

    public List<Occurrence> getResults(){
        return occurrences;
    }

    public LinkedList<ScraperLink> getPendingLinks() {
        return pendingLinks;
    }

    /**
     * Generates a report object that contains the relevant information of the scan.
     * @return ScraperReport object
     */
    public ScraperReport getReport(){
        ScraperReport report = new ScraperReport();
        report.setOccurrences(occurrences);
        report.setBaseUrl(baseUrl);
        report.setDepth(depth);
        report.setSearchTerms(searchTerms);
        pendingLinks = null;
        scannedLinks = null;
        return report;
    }

    /***PUBLISHER METHODS***/
    /**
     * Publishes changes on current URL being scanned
     * @param currentUrl
     * @param currentDepth
     */
    private void publishCurrentUrlChanged(String currentUrl, int currentDepth) {
        for(ScraperStatusListener l : listeners)
            l.onCurrentUrlChanged(currentUrl+"("+currentDepth+")");
    }

    /**
     * Publishes errors on specific Urls
     * @param currentUrl
     * @param message Error message
     */
    private void publishUrlError(String currentUrl, String message) {
        for(ScraperStatusListener l : listeners)
            l.onUrlError(currentUrl,message);
    }

    /**
     * Publishes a new link has been found
     * @param href URL of the link
     * @param size Total number of pending links
     */
    private void publishFoundNewLink(String href, int size) {
        for(ScraperStatusListener l : listeners)
            l.onNewLinkFound(href,size);
    }

    /**
     * Publishes the scan for a specific link has been finished
     * @param currentUrl
     */
    private void publishFinishedScanningUrl(String currentUrl) {
        for(ScraperStatusListener l : listeners)
            l.onFinishedScanningUrl(currentUrl,scannedLinks.size());
    }

    /**
     * Publishes the waiting status of the crawler
     * @param b
     */
    private void publishWaitStatus(boolean b) {
        for(ScraperStatusListener l : listeners)
            l.onWaitStatusChanged(b);
    }

    /**
     * Publishes the end of the scan
     */
    private void publishScrapeFinished() {
        for(ScraperStatusListener l : listeners)
            l.onScanFinished();
    }


}
